#  Copyright (c) 2022 by Amplo.

"""
Feature processor for extracting no features at all.
"""

from amplo.automl.feature_processing._base import BaseFeatureExtractor

__all__ = ["NopFeatureExtractor"]


class NopFeatureExtractor(BaseFeatureExtractor):
    """
    Feature processor for extracting no features.

    Each input column will be accepted as a feature.
    """

    _feature_translation = [(".*", None, None)]

    def _fit_transform(self, x, y=None, **fit_params):
        self.add_features(x)
        return self._transform(x=x, y=y)

    def _transform(self, x, y=None):
        return x[self.features_]



#  Copyright (c) 2022 by Amplo.

import json
from pathlib import Path

import joblib
import numpy as np
import pandas as pd
import scipy.stats

__all__ = [
    "DummyDataSampler",
    "make_data",
    "make_cat_data",
    "make_num_data",
    "make_interval_data",
    "make_production_data",
]


class DummyDataSampler:
    def __init__(self, distribution=scipy.stats.norm):
        """
        Class for sampling from a given distribution

        Parameters
        ----------
        distribution : scipy.stats.rv_continuous
            Distribution to sample from
        """
        self.transform = distribution

    def sample_data(self, num_samples: int):
        """Sample data"""
        cum = np.random.uniform(0.01, 0.99, num_samples)
        data = self.transform.ppf(cum)
        return data


def make_data(num_samples, *, cat_choices=None, num_dists=None):
    """
    Randomly sample categorical and/or numerical dummy data

    Parameters
    ----------
    num_samples : int
        Number of samples
    cat_choices : list of str or list of list or bool, optional
        (categorical) Specifies the choices to sample from
    num_dists : str or list of str or bool, optional
        (numerical) Specifies the distributions to sample from

    Returns
    -------
    tuple (pd.DataFrame, dict of {str : list})
    """

    # Handle input
    def parse_input(input_, default_value, list_of_list=False, list_of_str=False):
        if input_ is None or input_ is True:
            return default_value
        elif not input_:
            return []  # leave it empty
        elif (list_of_list and not isinstance(input_[0], list)) or (
            list_of_str and not isinstance(input_, list)
        ):
            return [input_]
        return input_

    cat_choices = parse_input(
        cat_choices, [list("abc"), list("xyz")], list_of_list=True
    )
    num_dists = parse_input(
        num_dists,
        ["uniform::0::100", "norm", "expon::0.4", "gamma::0.3", "beta::0.2::0.4"],
        list_of_str=True,
    )
    assert any(
        [cat_choices, num_dists]
    ), "Please specify at least one categorical or numerical column"

    # Sample categorical data
    cat_df = pd.DataFrame()
    for i, choice in enumerate(cat_choices):
        cat_df[f"cat_{i}"] = np.random.choice(choice, (num_samples,))

    # Sample numerical data
    num_df = pd.DataFrame()
    for i, dist in enumerate(num_dists):
        # Convert string to function
        if isinstance(dist, str):
            splits = dist.split("::")
            dist = splits[0]
            args = [float(arg) for arg in splits[1:]]
            dist = getattr(scipy.stats, dist)(*args)
        # Add distribution to corresponding array
        num_df[f"num_{i}"] = DummyDataSampler(dist).sample_data(num_samples)

    # Concatenate all data (which is not empty)
    df = pd.concat([df_ for df_ in (cat_df, num_df) if not df_.empty], axis=1)
    info = dict(cat_cols=cat_df.columns, num_cols=num_df.columns)

    return df, info


def make_cat_data(num_samples: int, cat_choices=None):
    return make_data(num_samples, cat_choices=cat_choices, num_dists=False)


def make_num_data(num_samples: int, num_dists=None):
    return make_data(num_samples, cat_choices=False, num_dists=num_dists)


def make_interval_data(n_logs=2, n_labels=2, directory=None, target="labels", **kwargs):
    """
    Create dummy data for Amplo`s IntervalAnalyser

    Parameters
    ----------
    n_logs : int
        Number of dummy logs / log files to create
    n_labels : int
        Number of dummy labels per log file
    directory : str or Path, optional
        Parent directory where interval data will be stored.
        If no directory is provided it will return a pandas.DataFrame instead.
    target : str
        Name of target column
    **kwargs
        Keyword arguments will be passed to ``make_data`` function

    Returns
    -------
    pd.DataFrame, optional
        Multi-indexed DataFrame such as the IntervalAnalyser would return it
    """

    # Define log names
    log_names = [f"DummyLog_{i}" for i in range(n_logs)]
    label_names = [f"DummyLabel_{i}" for i in range(n_labels)]
    # List of dataframes
    dfs = []

    for log in log_names:
        for label in label_names:

            num_samples = 20
            x, _ = make_data(num_samples, **kwargs)

            if directory:
                # Save in folder / label / log.csv
                save_dir = Path(directory) / label
                save_dir.mkdir(parents=True, exist_ok=True)
                x.to_csv(save_dir / f"{log}.csv", index=False)
            else:
                # Set multi-index
                index = pd.MultiIndex.from_arrays(
                    [num_samples * [log], range(num_samples)], names=["log", "index"]
                )
                x.set_index(index, inplace=True)
                # Add label column
                x[target] = label
                # Append to list
                dfs += [x]

    if not directory:
        return pd.concat(dfs)


def make_production_data(
    main_dir,
    team="DummyTeam",
    machine="DummyMachine",
    service="DummyService",
    issue="DummyIssue",
    version=1,
):
    """
    Create dummy production data, simulating output of the Amplo`s Pipeline

    Parameters
    ----------
    main_dir : str or Path
    team : str
    machine : str
    service : str
    issue : str
    version : int

    Returns
    -------
    tuple of Path, dict
        Model issue path AND dictionary containing info about dummy model
    """
    # Setup
    issue_dir = Path(main_dir) / team / machine / service / "models" / issue
    production_v_dir = issue_dir / "Production" / f"v{version}"
    production_v_dir.mkdir(parents=True, exist_ok=True)

    # Save dummy settings
    json.dump(dict(), open(production_v_dir / "Settings.json", "w"))
    # Save dummy model
    joblib.dump("", production_v_dir / "Model.joblib")
    # Save dummy report
    (production_v_dir / "Report.pdf").touch()

    return issue_dir, dict(
        team=team, machine=machine, service=service, issue=issue, version=version
    )



#  Copyright (c) 2022 by Amplo.

"""
Feature processor for extracting static features.
"""

import json
import re
from warnings import warn

import numpy as np
import pandas as pd
from sklearn.cluster import MiniBatchKMeans
from tqdm import tqdm

from amplo.automl.feature_processing._base import (
    BaseFeatureExtractor,
    sanitize_dataframe,
)

__all__ = ["StaticFeatureExtractor"]


class StaticFeatureExtractor(BaseFeatureExtractor):
    """
    Feature extractor for static data.

    Parameters
    ----------
    mode : str
        Model mode: {"classification", "regression"}.
    verbose : int
        Verbosity for logger.
    """

    _add_to_settings = [
        "means_",
        "stds_",
        "centers_",
        *BaseFeatureExtractor._add_to_settings,
    ]
    _feature_translation = [
        ("^((?!__).)*$", None, None),
        ("__(mul|div|x|d)__", "__(.)+__", "left-and-right"),
        ("^(dist|sin|cos|inv)__", "(dist|sin|cos|inv)__", "right"),
    ]

    def __init__(self, mode="notset", verbose=0):
        super().__init__(mode=mode, verbose=verbose)

    def _fit_transform(self, x, y=None, **fit_params):
        self.logger.info("Start fitting data.")

        # Input checks
        x, y = self._check_x_y(x, y)
        numeric_cols = [
            col for col, typ in zip(x, x.dtypes) if np.issubdtype(typ, np.number)
        ]
        if set(numeric_cols) != set(x):
            warn(
                "Handling non-numeric data is (currently) not supported. "
                "Corresponding columns will be ignored.",
                UserWarning,
            )
            x = x[numeric_cols]

        # Initialize fitting
        self._set_validation_model()
        self._init_feature_baseline_scores(x, y)

        # Fit features
        x_out = pd.concat(
            [
                self._fit_transform_raw_features(x),
                self._fit_transform_cross_features(x, y, update_baseline=True),
                self._fit_transform_k_means_features(x, y, update_baseline=True),
                self._fit_transform_trigo_features(x, y, update_baseline=True),
                self._fit_transform_inverse_features(x, y, update_baseline=True),
            ],
            axis=1,
        )

        self.logger.info("Finished fitting.")
        return sanitize_dataframe(x_out[self.features_])

    def _transform(self, x, y=None):
        self.logger.info("Transforming data.")

        # Handle input
        x = self._check_x(x)

        # Apply transformations
        x_out = pd.concat(
            [
                self._transform_raw_features(x),
                self._transform_cross_features(x),
                self._transform_k_means_features(x),
                self._transform_trigo_features(x),
                self._transform_inverse_features(x),
            ],
            axis=1,
        )

        return sanitize_dataframe(x_out[self.features_])

    # ----------------------------------------------------------------------
    # Feature processing

    @property
    def raw_features_(self):
        out = [str(c) for c in self.features_ if not re.search(".+__.+", c)]
        return sorted(out)

    def _fit_transform_raw_features(self, x, y=None):
        self.logger.info(f"Adding {x.shape[1]} raw features.")

        # Add accepted features
        self.add_features(x)

        return x

    def _transform_raw_features(self, x):
        if not self.raw_features_:
            self.logger.debug("No raw features added.")
            return pd.DataFrame(index=x.index)

        self.logger.info("Transforming raw features.")

        assert set(self.raw_features_).issubset(
            x
        ), "Expected raw features do not match with actual."
        x_out = x[self.raw_features_]

        return x_out

    @property
    def cross_features_(self):
        out = [str(c) for c in self.features_ if re.search("__(mul|div|x|d)__", c)]
        return sorted(out)

    def _fit_transform_cross_features(self, x, y, update_baseline=True):
        self.logger.info("Fitting cross features.")

        scores = {}
        x_out = []  # empty df for concat (later)
        for i, col_a in enumerate(tqdm(x)):
            col_a_useless_so_far = True
            for j, col_b in enumerate(x.iloc[:, i + 1 :]):
                # Skip when same column or `col_a` is potentially useless.
                if col_a == col_b or (
                    j > max(50, x.shape[0] // 3) and col_a_useless_so_far
                ):
                    continue

                # Make __div__ feature
                div_feature = x[col_a] / x[col_b].replace(0, 1e-10)
                div_score = self._calc_feature_scores(div_feature, y)
                if self.accept_feature(div_score):
                    col_a_useless_so_far = False
                    name = f"{col_a}__div__{col_b}"
                    scores[name] = div_score
                    x_out += [div_feature.rename(name)]

                # Make __mul__ feature
                mul_feature = x[col_a] * x[col_b]
                mul_score = self._calc_feature_scores(mul_feature, y)
                if self.accept_feature(mul_score):
                    name = "{}__mul__{}".format(*sorted([col_a, col_b]))
                    col_a_useless_so_far = False
                    scores[name] = mul_score
                    x_out += [mul_feature.rename(name)]

        # Decide which features to accept
        scores = self.select_scores(
            pd.DataFrame(scores), best_n_per_class=50, update_baseline=update_baseline
        )
        x_out = (
            pd.concat(x_out, axis=1)[scores.columns]
            if x_out
            else pd.DataFrame(index=x.index)
        )
        self.logger.info(f"Accepted {x_out.shape[1]} cross features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_cross_features(self, x):
        if not self.cross_features_:
            self.logger.debug("No cross features added.")
            return pd.DataFrame(index=x.index)

        self.logger.info("Transforming cross features.")

        x_out = []
        for feature_name in self.cross_features_:
            # Deprecation support
            if "__x__" in feature_name:
                col_a, col_b = feature_name.split("__x__")
                feat = x[col_a] * x[col_b]
                x_out += [feat.rename(feature_name)]
            elif "__d__" in feature_name:
                col_a, col_b = feature_name.split("__d__")
                feat = x[col_a] / x[col_b].replace(0, -1e-10)
                x_out += [feat.rename(feature_name)]
            # New names
            elif "__mul__" in feature_name:
                col_a, col_b = feature_name.split("__mul__")
                feat = x[col_a] * x[col_b]
                x_out += [feat.rename(feature_name)]
            elif "__div__" in feature_name:
                col_a, col_b = feature_name.split("__div__")
                feat = x[col_a] / x[col_b].replace(0, -1e-10)
                x_out += [feat.rename(feature_name)]
            else:
                raise ValueError(f"Cross feature not recognized: {feature_name}")

        x_out = pd.concat(x_out, axis=1)

        assert set(self.cross_features_) == set(
            x_out
        ), "Expected cross features do not match with actual."

        return x_out

    @property
    def k_means_features_(self):
        out = [str(c) for c in self.features_ if re.match("dist__", c)]
        return sorted(out)

    def _fit_transform_k_means_features(self, x, y, update_baseline=True):
        self.logger.info("Fitting k-Means features.")

        # Prepare data
        means, stds = x.mean(), x.std().replace(0, 1)
        self.means_ = json.loads(means.to_json())
        self.stds_ = json.loads(stds.to_json())
        x = (x - means) / stds

        # Determine clusters
        n_clusters = min(x.shape[1], max(8, int(np.log10(x.shape[1] * 8))))
        k_means = MiniBatchKMeans(n_clusters=n_clusters)
        col_names = [f"dist__{c}_{n_clusters}" for c in range(n_clusters)]
        distances = pd.DataFrame(
            k_means.fit_transform(x), columns=col_names, index=x.index
        )
        distances = sanitize_dataframe(distances)
        centers = pd.DataFrame(k_means.cluster_centers_, columns=x.columns)
        self.centers_ = json.loads(centers.to_json())

        # Score and decide which features to accept
        scores = self.select_scores(
            distances.apply(self._calc_feature_scores, y=y),
            best_n_per_class=50,
            update_baseline=update_baseline,
        )
        x_out = distances[scores.columns]
        self.logger.info(f"Accepted {x_out.shape[1]} k-Means features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_k_means_features(self, x):
        if not self.k_means_features_:
            self.logger.debug("No k-Means features added.")
            return pd.DataFrame(index=x.index)

        self.logger.info("Transforming k-Means features.")

        # Init
        x = (x - pd.Series(self.means_)) / pd.Series(self.stds_)  # normalize
        centers = pd.DataFrame(self.centers_)

        # Extract features
        x_out = []
        for feature_name in self.k_means_features_:
            cluster_idx, _ = feature_name[len("dist__") :].split("_")  # remove prefix
            cluster_center = centers.iloc[int(cluster_idx)]
            feat = ((x - cluster_center) ** 2).sum(1) ** 0.5
            x_out += [feat.rename(feature_name)]
        x_out = pd.concat(x_out, axis=1)

        assert set(self.k_means_features_) == set(
            x_out
        ), "Expected k-Means features do not match with actual."

        return x_out

    @property
    def trigo_features_(self):
        out = [str(c) for c in self.features_ if re.match("(sin|cos)__", c)]
        return sorted(out)

    def _fit_transform_trigo_features(self, x, y, update_baseline=True):
        self.logger.info("Fitting trigonometric features.")

        # Make features
        sin_x = np.sin(x).rename(columns={col: f"sin__{col}" for col in x})
        cos_x = np.cos(x).rename(columns={col: f"cos__{col}" for col in x})
        feats = pd.concat([sin_x, cos_x], axis=1)

        # Score and decide which features to accept
        scores = self.select_scores(
            feats.apply(self._calc_feature_scores, y=y, axis=0),
            best_n_per_class=50,
            update_baseline=update_baseline,
        )
        x_out = feats[scores.columns]
        self.logger.info(f"Accepted {x_out.shape[1]} raw features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_trigo_features(self, x):
        if not self.trigo_features_:
            self.logger.debug("No trigonometric features added.")
            return pd.DataFrame(index=x.index)

        self.logger.info("Transforming trigonometric features.")

        # Group by transformation
        feat_info = [list(f.partition("__"))[::2] for f in self.trigo_features_]
        feat_info = pd.DataFrame(feat_info).groupby(0).agg(list)[1]

        # Transform
        x_out = []
        for func, cols in feat_info.iteritems():
            col_names = {col: f"{func}__{col}" for col in x}
            x_out += [getattr(np, func)(x[cols]).rename(columns=col_names)]
        x_out = pd.concat(x_out, axis=1)

        assert set(self.trigo_features_) == set(
            x_out
        ), "Expected trigonometric features do not match with actual."

        return x_out

    @property
    def inverse_features_(self):
        out = [str(c) for c in self.features_ if re.match("inv__", c)]
        return sorted(out)

    def _fit_transform_inverse_features(self, x, y, update_baseline=True):
        self.logger.info("Fitting inverse features.")

        # Make features
        with np.errstate(divide="ignore"):  # ignore true_divide warnings
            feats = (1.0 / x).rename(columns={col: f"inv__{col}" for col in x})
        feats = sanitize_dataframe(feats)  # remove np.inf for scoring

        # Score and decide which features to accept
        scores = self.select_scores(
            feats.apply(self._calc_feature_scores, y=y, axis=0),
            best_n_per_class=50,
            update_baseline=update_baseline,
        )
        x_out = feats[scores.columns]
        self.logger.info(f"Accepted {x_out.shape[1]} inverse features.")

        # Add accepted features
        self.add_features(x_out)

        return x_out

    def _transform_inverse_features(self, x):
        if not self.inverse_features_:
            self.logger.debug("No inverse features added.")
            return pd.DataFrame(index=x.index)

        self.logger.info("Transforming inverse features.")

        # Get all columns to invert
        inv_columns = [
            f[len("inv__") :] for f in self.inverse_features_  # remove prefix
        ]

        # Transform
        with np.errstate(divide="ignore"):  # ignore true_divide warnings
            x_out = (1.0 / x[inv_columns]).rename(
                columns={col: f"inv__{col}" for col in x}
            )

        assert set(self.inverse_features_) == set(
            x_out
        ), "Expected inverse features do not match with actual."

        return x_out



#  Copyright (c) 2022 by Amplo.

from pathlib import Path

import numpy as np
import pandas as pd

from amplo.automl import IntervalAnalyser
from amplo.utils.testing import make_interval_data
from tests import rmtree


def create_data_frames(n_samples, n_features):
    dim = (int(n_samples / 2), n_features)
    columns = [f"Feature_{i}" for i in range(n_features)]
    df1 = pd.DataFrame(
        columns=columns,
        data=np.vstack((np.random.normal(0, 1, dim), np.random.normal(100, 1, dim))),
    )
    df2 = pd.DataFrame(
        columns=columns,
        data=np.vstack((np.random.normal(0, 1, dim), np.random.normal(-100, 1, dim))),
    )
    return df1, df2


def create_test_folders(directory: Path, n_samples, n_features):
    # Make directories
    for sub_folder in ("Class_1", "Class_2"):
        (directory / sub_folder).mkdir(exist_ok=True, parents=True)

    # Create and save dataframes
    for i in range(140):
        df1, df2 = create_data_frames(n_samples, n_features)
        df1.to_csv(directory / f"Class_1/Log_{i}.csv", index=False)
        df2.to_csv(directory / f"Class_2/Log_{i}.csv", index=False)


class TestIntervalAnalyser:
    @classmethod
    def setup_class(cls):
        cls.n_samples = 50
        cls.n_features = 25
        cls.ia_folder = Path("IA")
        rmtree(cls.ia_folder)
        create_test_folders(cls.ia_folder, cls.n_samples, cls.n_features)

    @classmethod
    def teardown_class(cls):
        rmtree(cls.ia_folder)

    def test_all(self):
        # Set up IntervalAnalyser
        ia = IntervalAnalyser(min_length=10)
        ia.fit_transform(self.ia_folder)

        # Attribute tests
        assert ia.n_folders == 2
        assert ia.n_files == 280

        # Functional tests
        for i in range(ia.n_files):
            dist = ia._distributions[i].values
            assert all(
                v < 0.8 for v in dist[: int(len(dist) / 2)]
            ), "Noise with high percentage of neighbors"
            assert all(
                v > 0.8 for v in dist[int(len(dist) / 2) :]
            ), "Information with low percentage of neighbors"

        # Data tests
        data_no_noise = ia.data_without_noise
        data_full = ia.data_with_noise
        assert (
            len(data_full) == ia.n_samples and len(data_no_noise) == ia.n_samples // 2
        ), "Incorrect number of samples"
        assert (
            data_full.index.get_level_values(0).nunique() == ia.n_files
        ), "Files skipped"
        assert (
            len(ia._features.columns) <= self.n_features
        ), "Incorrect number of features"

    def test_parse_data(self):
        # Test receiving dataframe and receiving directory
        for test in ("dataframe", "single_dataframe", "directory"):
            ia = IntervalAnalyser(target="labels")

            if test == "dataframe":
                # Create dummy data
                features = make_interval_data(
                    directory=None, target=ia.target, cat_choices=False
                )
                labels = features.pop(ia.target)
                # Set parse arguments
                parse_args = (features, labels)
            elif test == "single_dataframe":
                # Create dummy data
                data = make_interval_data(
                    directory=None, target=ia.target, cat_choices=False
                )
                # Set parse arguments
                parse_args = (data,)
            elif test == "directory":
                # Create dummy data
                make_interval_data(
                    directory=self.ia_folder, target=ia.target, cat_choices=False
                )
                # Set parse arguments
                parse_args = (self.ia_folder, None)
            else:
                raise NotImplementedError()

            # Read data
            _, _ = ia._parse_data(*parse_args)

            # Data checks
            assert all(
                isinstance(data, (pd.Series, pd.DataFrame))
                for data in (ia._labels, ia._features, ia.orig_data)
            ), "Data should not be empty"
            assert (
                len(ia._features.select_dtypes(include=["datetime64"]).columns) == 0
            ), "Date-time columns should have been eliminated, internally"
            assert (
                len(ia._features.select_dtypes(include=["float64"]).columns) == 0
            ), "Dtype float64 columns should have been converted to float32"

            # Other checks
            assert not ia.is_fitted, "IntervalAnalyser should not yet be fitted"
            assert ia._noise_indices is None, "No noise indices should be set yet"

    # TODO test correlation warning
    # TODO test data quality warning
    # TODO test multi-index error
    # TODO test index misalignment error
    # TODO test feature/sample length error
    # TODO test target in features error
    # TODO test target != labels.name



#  Copyright (c) 2022 by Amplo.

import pytest
from sklearn.datasets import load_iris, make_regression

from amplo.automl import DataExplorer
from tests import rmtree


@pytest.fixture(scope="class", params=["classification", "regression"])
def make_mode(request):
    mode = request.param
    if mode == "classification":
        x, y = load_iris(return_X_y=True, as_frame=True)
    elif mode == "regression":
        # Interesting enough, the regression test for `test_explorer` ends up in a
        # deadlock, specifically in the `DataExplorer.shap()` method. It's even more
        # interesting that the california housing dataset works fine when using
        # `unittest`
        # x, y = fetch_california_housing(return_X_y=True, as_frame=True)
        x, y = make_regression()
    else:
        raise NotImplementedError("Invalid mode")
    request.cls.x = x
    request.cls.y = y
    request.cls.mode = mode
    yield


@pytest.mark.usefixtures("make_mode")
class TestDataExploring:
    def test_explorer(self):
        tmp_folder = "tmp/"
        eda = DataExplorer(self.x, y=self.y, mode=self.mode, folder=tmp_folder)
        eda.run()
        rmtree(tmp_folder)


# class TestDataExploring(unittest.TestCase):
#
#     @classmethod
#     def setUpClass(cls):
#         cls.class_x, cls.class_y = load_iris(return_X_y=True, as_frame=True)
#         cls.reg_x, cls.reg_y = fetch_california_housing(return_X_y=True)
#
#     def test_regression(self):
#         eda = DataExplorer(self.reg_x, y=self.reg_y, mode='regression', folder='tmp/')
#         eda.run()
#         rmtree('tmp/')
#
#     def test_classification(self):
#         eda = DataExplorer(self.class_x, y=self.class_y, mode='classification', folder='tmp/')  # noqa: E501
#         eda.run()
#         rmtree('tmp/')



#  Copyright (c) 2022 by Amplo.

"""
Feature processor for extracting and selecting features.
"""
from __future__ import annotations

import re

import numpy as np
import pandas as pd
import psutil
from shap import TreeExplainer
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from amplo.automl.feature_processing._base import (
    BaseFeatureExtractor,
    BaseFeatureProcessor,
    sanitize_series,
)
from amplo.automl.feature_processing.nop_feature_extractor import NopFeatureExtractor
from amplo.automl.feature_processing.static_feature_extractor import (
    StaticFeatureExtractor,
)
from amplo.automl.feature_processing.temporal_feature_extractor import (
    TemporalFeatureExtractor,
)
from amplo.base.exceptions import NotFittedError
from amplo.classification import CatBoostClassifier
from amplo.regression import CatBoostRegressor
from amplo.utils import check_dtypes

__all__ = ["find_collinear_columns", "FeatureProcessor"]


def find_collinear_columns(data, information_threshold=0.9):
    """
    Finds collinear features and returns them.

    Calculates the Pearson Correlation coefficient for all input features.
    Features that exceed the information threshold are considered linearly
    co-dependent, i.e. describable by: y = a * x + b. As these features add
    little to no information, they will be removed.

    Parameters
    ----------
    data : pd.DataFrame
        Data to search for collinear features.
    information_threshold : float
        Percentage value that defines the threshold for a ``collinear`` feature.

    Returns
    -------
    list of str
        List of collinear feature columns.
    """
    if not isinstance(data, pd.DataFrame):
        raise ValueError(f"Invalid dtype for `data`: {type(data).__name__}")
    if not isinstance(information_threshold, float):
        raise ValueError(
            f"Invalid dtype for `threshold`: {type(information_threshold).__name__}"
        )
    if not 0 < information_threshold < 1:
        raise ValueError(
            "`information_threshold` must be a valid percentage, "
            "excluding 0 and 1, i.e., (0, 1)."
        )

    # Get collinear features
    nk = data.shape[1]
    corr_mat = np.zeros((nk, nk))

    try:
        # Check available memory and raise error if necessary
        mem_avail = psutil.virtual_memory().available
        mem_data = data.memory_usage(deep=True).sum()
        if mem_avail < 2 * mem_data:
            raise MemoryError(
                "Data is too big to handle time efficient. Using memory efficient "
                "implementation instead."
            )

        # More efficient in terms of time but may crash when data size is huge
        norm_data = (data - data.mean(skipna=True, numeric_only=True)).to_numpy()
        ss = np.sqrt(np.sum(norm_data**2, axis=0))

        for i in range(nk):
            for j in range(nk):
                if i >= j:
                    continue
                sum_ = np.sum(norm_data[:, i] * norm_data[:, j])
                corr_mat[i, j] = abs(sum_ / (ss[i] * ss[j]))

    except MemoryError:
        # More redundant calculations but more memory efficient
        for i, col_name_i in enumerate(data):
            col_i = data[col_name_i]
            norm_col_i = (col_i - col_i.mean(skipna=True)).to_numpy()
            del col_i
            ss_i = np.sqrt(np.sum(norm_col_i**2))

            for j, col_name_j in enumerate(data):
                if i >= j:
                    continue

                col_j = data[col_name_j]
                norm_col_j = (col_j - col_j.mean(skipna=True)).to_numpy()
                del col_j
                ss_j = np.sqrt(np.sum(norm_col_j**2))

                sum_ = np.sum(norm_col_i * norm_col_j)
                corr_mat[i, j] = abs(sum_ / (ss_i * ss_j))

    # Set collinear columns
    mask = np.sum(corr_mat > information_threshold, axis=0) > 0
    collinear_columns = np.array(data.columns)[mask].astype(str).tolist()
    return collinear_columns


class FeatureProcessor(BaseFeatureProcessor):
    """
    Feature processor module to extract and select features.

    Parameters
    ----------
    mode : {"notset", "classification", "regression"}, optional, default: "notset"
        Model mode.
    is_temporal : bool, optional
        Whether the data should be treated as temporal data or not.
        If none is provided, is_temporal will be set to true when fit data is
        multi-indexed, false otherwise.
    extract_features : bool
        Whether to extract features or just remove correlating columns.
    collinear_threshold : float
        Information threshold for collinear features.
    analyse_feature_sets : {"auto", "all", "gini", "shap"}, optional, default: "auto"
        Which feature sets to analyse.
        If None, no feature set will be analysed.
        If "auto", gini (and shap) will be analysed.
        If "all", gini and shap will be analysed.
        If "gini" or "shap", gini or shap will be analysed, respectively.
    selection_cutoff : float
        Upper feature importance threshold for threshold feature selection.
    selection_increment : float
        Lower feature importance threshold for increment feature selection.
    verbose : int
        Verbosity for logger.
    **extractor_kwargs : typing.Any
        Additional keyword arguments for feature extractor.
        Currently, only the `TemporalFeatureExtractor` module supports this parameter.
    """

    _add_to_settings = [
        *BaseFeatureProcessor._add_to_settings,
        "datetime_cols_",
        "collinear_cols_",
        "numeric_cols_",
        "feature_extractor",
        "feature_importance_",
        "feature_sets_",
    ]

    def __init__(
        self,
        mode="notset",
        is_temporal=None,
        extract_features=True,
        collinear_threshold=0.99,
        analyse_feature_sets="auto",
        selection_cutoff=0.85,
        selection_increment=0.005,
        verbose=1,
        **extractor_kwargs,
    ):
        super().__init__(mode=mode, verbose=verbose)

        check_dtypes(
            ("is_temporal", is_temporal, (bool, type(None))),
            ("extract_features", extract_features, bool),
            ("collinear_threshold", collinear_threshold, float),
            ("analyse_feature_sets", analyse_feature_sets, (str, type(None))),
            ("selection_cutoff", selection_cutoff, float),
            ("selection_increment", selection_increment, float),
        )
        for value, name in (
            (collinear_threshold, "collinear_threshold"),
            (selection_cutoff, "selection_cutoff"),
            (selection_increment, "selection_increment"),
        ):
            if not 0 < value < 1:
                raise ValueError(f"Invalid argument {name} = {value} ∉ (0, 1).")

        # Set attributes
        self.is_temporal = is_temporal
        self.extract_features = extract_features
        self.collinear_threshold = collinear_threshold
        self.analyse_feature_sets = analyse_feature_sets
        self.selection_cutoff = selection_cutoff
        self.selection_increment = selection_increment
        self.feature_extractor = None
        self.extractor_kwargs = extractor_kwargs

    def _fit_transform(self, x, y=None, **fit_params):
        self.logger.info("Start fitting data.")

        # Input checks
        x, y = self._check_x_y(x, y)

        # Define which columns are datetime, numeric and collinear.
        self._find_columns_of_interest(x)
        x = x[self.numeric_cols_]

        # Fit and transform feature extractor.
        self._check_is_temporal_attribute(x)
        x_ext = self._fit_transform_feature_extractor(x, y)
        y_ext = self.transform_target(y)

        # Analyse feature importance and feature sets
        self._analyse_feature_sets(x_ext, y_ext)

        self.logger.info("Finished fitting.")
        return x_ext

    def transform(self, x, y=None, feature_set=None):
        """
        Transform data and return it.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_".
            self._is_fitted

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Feature data to transform.
        y : numpy.ndarray or pandas.Series
            Additional target data to transform.
        feature_set : str, optional
            Desired feature set.
            When feature_set is None, all features will be returned.

        Returns
        -------
        pandas.DataFrame
            Transformed version of x.
        """
        # Check whether is fitted
        self.check_is_fitted()

        # Set features for transformation
        if feature_set is None:
            features = self.features_
        elif feature_set in self.feature_sets_:
            features = self.feature_sets_[feature_set]
        else:
            raise ValueError(f"Feature set does not exist: {feature_set}")

        # Remove features for faster transforming.
        # We (temporarily) remove features from the feature extractor for speed-up.
        # Only `features_` will be extracted.
        orig_features_ = self.features_
        self.features_ = features

        # Transform data
        xt = self._transform(x, y)

        # Restore original features
        self.features_ = orig_features_

        return xt

    def _transform(self, x, y=None, feature_set=None):
        self.logger.info("Transforming data.")

        # Handle input
        x = self._check_x(x)
        x = self._impute_missing_columns(x)

        # Transform
        return self.feature_extractor.transform(x)

    def transform_target(self, y):
        """
        Transform target column (only if temporal data).

        Parameters
        ----------
        y : pd.Series

        Returns
        -------
        pd.Series
        """
        if isinstance(self.feature_extractor, TemporalFeatureExtractor):
            y = self.feature_extractor._fit_data_to_window_size(y)  # noqa
            y = self.feature_extractor._pool_target(y)  # noqa

        return y

    # ----------------------------------------------------------------------
    # Feature name translation

    def translate_features(self, feature_columns: list[str]) -> dict[str, list[str]]:
        """
        Translates (extracted) features and tells its underlying original feature.

        Parameters
        ----------
        feature_columns : list of str
            Feature columns to be translated.

        Returns
        -------
        dict of {str: list of str}
            Dictionary with `feature_columns` as keys and their underlying original
            features as values.
        """
        check_dtypes(("item", item, str) for item in feature_columns)

        translation = {}
        for filtr, split, pos in self.feature_extractor._feature_translation:  # noqa
            for col in feature_columns:
                if not re.search(filtr, col):
                    # Column doesn't match the filter
                    continue
                elif col in translation:
                    # We don't want to overwrite an existing key-value pair
                    raise ValueError(
                        "`_feature_translation` doesn't give unique filters."
                    )
                elif split is None:
                    translation[col] = [col]  # 1-to-1 translation
                    continue

                left, _, right = re.split(split, col, maxsplit=1)
                if pos == "left":
                    translation[col] = [left]
                elif pos == "right":
                    translation[col] = [right]
                elif pos == "left-and-right":
                    translation[col] = [left, right]
                else:
                    raise ValueError(f"Unknown positioning identifier: {pos}")

        if set(translation) != set(feature_columns):
            raise ValueError(
                f"Some feature columns are unknown how to handle: "
                f"{set(feature_columns) - set(translation)}"
            )

        return translation

    def get_required_columns(self, feature_set=None):
        """
        Returns all required data columns that are required.

        Parameters
        ----------
        feature_set : str, optional
            Desired feature set.
            When no feature_set is `None` all features will be returned.

        Returns
        -------
        list of str
            All required feature columns for given feature set.
        """
        # Set features for transformation
        if feature_set is None:
            features = self.features_
        elif feature_set in self.feature_sets_:
            features = self.feature_sets_[feature_set]
        else:
            raise ValueError(f"Feature set does not exist: {feature_set}")

        # A feature name always contains before and/or after two underscores.
        # Based on this assumption we use RegEx to find all features that are
        # required for the given feature set (of feature names).
        all_cols = set(self.datetime_cols_ + self.numeric_cols_)
        required_cols = sorted(
            set(
                [
                    col
                    for col in all_cols
                    if any(re.search(f"(__{col}|{col}__)", f) for f in features)
                ]
            )
        )

        return required_cols

    # ----------------------------------------------------------------------
    # Feature processing

    def _check_is_temporal_attribute(self, x):
        """
        Checks is_temporal attribute. If not set and x is multi-indexed, sets to true.

        Parameters
        ----------
        x : pd.DataFrame
        """
        # Check if `is_temporal` attribute is set
        if self.is_temporal is None:
            # When x is multi-indexed, we assume that we have temporal data.
            self.is_temporal = len(x.index.names) == 2

    def _find_columns_of_interest(self, x):
        """
        Examines the data and separates different column types.

        Fitted attributes:
            Datetime columns are stored in "datetime_cols_".
            Collinear, numeric columns are stored in "collinear_cols_".
            Numeric columns (not collinear) are stored in "numeric_cols_".

        Parameters
        ----------
        x : pd.DataFrame
            Data to examine.
        """
        self.logger.info("Analysing columns of interest.")

        self.datetime_cols_ = [
            col for col in x if pd.api.types.is_datetime64_any_dtype(x[col])
        ]
        non_datetime_cols = [col for col in x if col not in self.datetime_cols_]
        self.collinear_cols_ = find_collinear_columns(
            x.loc[:, non_datetime_cols], self.collinear_threshold
        )
        self.numeric_cols_ = [
            col for col in non_datetime_cols if col not in self.collinear_cols_
        ]

        self.logger.info(
            f"Found {len(self.datetime_cols_)} datetime columns and "
            f"{len(self.collinear_cols_) + len(self.numeric_cols_)} numeric "
            f"columns whereas {len(self.collinear_cols_)} are collinear and "
            f"thus removed."
        )

    def _impute_missing_columns(self, x):
        """
        Imputes missing columns when not present for transforming.

        Parameters
        ----------
        x : pd.DataFrame
            Data to check and impute when necessary.

        Returns
        -------
        pd.DataFrame
            Cleaned data.
        """
        self.check_is_fitted()

        # Find missing columns
        missing_datetime = [col for col in self.datetime_cols_ if col not in x]
        missing_numeric = [col for col in self.numeric_cols_ if col not in x]
        if missing_datetime or missing_numeric:
            self.logger.warning(
                f"Imputing {len(missing_datetime) + len(missing_numeric)} "
                f"missing columns."
            )

        # Impute missing datetime columns
        if missing_datetime:
            raise NotImplementedError(
                "Imputing missing datetime columns is " "currently not supported."
            )

        # Impute missing numeric columns with zeros
        for col in missing_numeric:
            x[col] = sanitize_series(pd.Series([0] * x.shape[0]))

        return x

    def _fit_transform_feature_extractor(self, x, y):
        """
        Selects feature processor instance and calls fit_transform.

        Parameters
        ----------
        x : pd.DataFrame
        y : pd.Series

        Returns
        -------
        pd.DataFrame
            Extracted features.
        """
        assert self.is_temporal is not None, "Forgot to set attribute."

        # Fit feature extractor
        if not self.extract_features:
            self.feature_extractor = NopFeatureExtractor(mode=None)
        elif self.is_temporal:
            self.feature_extractor = TemporalFeatureExtractor(
                mode=self.mode, verbose=self.verbose, **self.extractor_kwargs
            )
        else:
            self.feature_extractor = StaticFeatureExtractor(
                mode=self.mode, verbose=self.verbose  # , **self.extractor_kwargs
            )

        # Extract features
        x_ext = self.feature_extractor.fit_transform(x, y)

        # Find collinear features and remove those
        collinear_cols = find_collinear_columns(x_ext, self.collinear_threshold)
        self.feature_extractor.remove_features(collinear_cols)
        self.logger.info(
            f"Removed {len(collinear_cols)} collinear columns "
            f"from extracted features."
        )

        return x_ext[self.feature_extractor.features_]

    @property
    def features_(self):
        """
        Returns `features_` attribute of the feature extractor.

        Raises
        ------
        NotFittedError
            If the feature extractor is not yet set.
        """
        if not isinstance(self.feature_extractor, BaseFeatureExtractor):
            raise NotFittedError("Invalid feature extractor.")
        return self.feature_extractor.features_

    @features_.setter
    def features_(self, value):
        """
        Setter for `features_` attribute.

        Parameters
        ----------
        value : typing.List of str

        Raises
        ------
        NotFittedError
            If the feature extractor is not yet set.
        """
        if not isinstance(self.feature_extractor, BaseFeatureExtractor):
            raise NotFittedError("Invalid feature extractor.")
        self.feature_extractor.set_features(value)

    # ----------------------------------------------------------------------
    # Feature selection

    def _analyse_feature_sets(self, x, y):
        """
        Explores importance of features and defines selected subsets.

        Parameters
        ----------
        x : pd.DataFrame
        y : pd.Series

        Notes
        -----
        When ``analyse_feature_sets`` is false, ``feature_importance_`` is not set and
        ``feature_sets_`` has only one set, containing all columns.
        """
        # Init
        self.feature_importance_ = {}
        self.feature_sets_ = {}

        # Analyse
        analyse_gini = self.analyse_feature_sets in ("auto", "all", "gini")
        analyse_shap = self.analyse_feature_sets in ("all", "shap") or (
            self.analyse_feature_sets == "auto" and len(y) < 50_000
        )
        if analyse_gini:
            self._select_gini_impurity(x, y)
        if analyse_shap:
            self._select_shap(x, y)
        if not (analyse_gini or analyse_shap):
            self.feature_sets_["take_all"] = list(x.columns)

        # Enforce feature sets being sorted
        self.feature_sets_ = {
            key: sorted(values) for key, values in self.feature_sets_.items()
        }

    def _select_gini_impurity(self, x, y):
        """
        Selects features based on the random forest feature importance.

        Calculates the mean decrease in Gini impurity. Symmetric correlation
        based on multiple features and multiple tree ensembles.

        Parameters
        ----------
        x : pd.DataFrame
        y : pd.Series
        """
        self.logger.info("Analysing feature importance: Gini impurity.")

        # Set model
        rs = np.random.RandomState(seed=236868)
        if self.mode == "regression":
            forest = RandomForestRegressor(random_state=rs)
        elif self.mode == "classification" or self.mode == "multiclass":
            forest = RandomForestClassifier(random_state=rs)
        else:
            raise ValueError("Invalid mode.")
        forest.fit(x, y)

        # Get RF values
        fi = forest.feature_importances_
        fi_sum = np.sum(fi)
        idx_sort = np.flip(np.argsort(fi))

        # Set feature importance
        self.feature_importance_["rf"] = {
            col: importance
            for col, importance in zip(x.columns[idx_sort], fi[idx_sort])
        }

        # Threshold: Take best n columns to satisfy selection cutoff
        mask = fi[idx_sort].cumsum() - fi[idx_sort] <= fi_sum * self.selection_cutoff
        idx_keep = idx_sort[mask]
        threshold = x.columns[idx_keep].to_list()
        self.feature_sets_["rf_threshold"] = threshold
        self.logger.info(
            f"Selected {len(threshold)} features with "
            f"{self.selection_cutoff * 100:.2f}% RF threshold."
        )

        # Increment
        idx_keep = fi > fi_sum * self.selection_increment
        increment = x.columns[idx_keep].to_list()
        self.feature_sets_["rf_increment"] = increment
        self.logger.info(
            f"Selected {len(increment)} features with "
            f"{self.selection_increment * 100:.2f}% RF increment."
        )

    def _select_shap(self, x, y):
        """
        Calculates shapely value to be used as a measure of feature importance.

        Parameters
        ----------
        x : pd.DataFrame
        y : pd.Series
        """
        self.logger.info("Analysing feature importance: Shapely additive explanations.")

        # Set model
        seed = 236868
        if self.mode == "regression":
            base = CatBoostRegressor(random_seed=seed)
        elif self.mode == "classification" or self.mode == "multiclass":
            base = CatBoostClassifier(random_seed=seed)
        else:
            raise ValueError("Invalid mode.")
        base.fit(x, y)

        # Get Shap values
        explainer = TreeExplainer(base.model)
        shap = np.array(explainer.shap_values(x, y))

        # Average over classes if necessary
        if shap.ndim == 3:
            shap = np.mean(np.abs(shap), axis=0)

        # Average over samples
        shap = np.mean(np.abs(shap), axis=0)
        shap /= shap.sum()  # normalize
        shap_sum = np.sum(shap)
        idx_sort = np.flip(np.argsort(shap))

        # Add to class attribute
        self.feature_importance_["shap"] = {
            col: importance
            for col, importance in zip(x.columns[idx_sort], shap[idx_sort])
        }

        # Threshold
        mask = (
            shap[idx_sort].cumsum() - shap[idx_sort] <= shap_sum * self.selection_cutoff
        )
        idx_keep = idx_sort[mask]
        threshold = x.columns[idx_keep].to_list()
        self.feature_sets_["shap_threshold"] = threshold
        self.logger.info(
            f"Selected {len(threshold)} features with "
            f"{self.selection_cutoff * 100:.2f}% Shap threshold."
        )

        # Increment
        idx_keep = shap > shap_sum * self.selection_increment
        increment = x.columns[idx_keep].to_list()
        self.feature_sets_["shap_increment"] = increment
        self.logger.info(
            f"Selected {len(increment)} features with "
            f"{self.selection_increment * 100:.2f}% Shap increment."
        )



#  Copyright (c) 2022 by Amplo.

from amplo.documenting.binary import BinaryDocumenting
from amplo.documenting.multiclass import MultiDocumenting
from amplo.documenting.regression import RegressionDocumenting



#  Copyright (c) 2022 by Amplo.

"""
Base exceptions.
"""

__all__ = [
    "ExperimentalWarning",
    "NotFittedError",
]


class ExperimentalWarning(UserWarning):
    """
    Warning for experimental features.
    """


class NotFittedError(ValueError, AttributeError):
    """
    Exception class to raise if estimator is used before fitting.
    """



#  Copyright (c) 2022 by Amplo.

from typing import TypeVar

from amplo.grid_search.exhaustive import ExhaustiveGridSearch
from amplo.grid_search.halving import HalvingGridSearch
from amplo.grid_search.optuna import OptunaGridSearch

GridSearchType = TypeVar(
    "GridSearchType", ExhaustiveGridSearch, HalvingGridSearch, OptunaGridSearch
)



#  Copyright (c) 2022 by Amplo.

import numpy as np
import pandas as pd

from amplo.automl import Sequencer


class TestSequence:
    def test_init(self):
        assert Sequencer(), "Class initiation failed"

    def test_numpy_none(self):
        # Parameters
        features = 5
        length = 500
        back = np.random.randint(1, 50)
        x, y = np.random.randint(0, 50, (length, features)), np.random.randint(
            0, 50, length
        )

        # Iterate scenarios
        for forward in [[1], [5]]:
            # Sequence
            sequence = Sequencer(back=back, forward=forward)
            seq_x, seq_y = sequence.convert(x, y, flat=False)

            # Test
            assert len(seq_x.shape) == 3, "seq_x is no tensor"
            assert seq_x.shape[1] == back, "seq_x step dimension incorrect"
            assert seq_x.shape[2] == features, "seq_x feature dimension incorrect"
            assert (
                seq_x.shape[0] == seq_y.shape[0]
            ), "seq_x and seq_y have inconsistent samples"
            for i in range(seq_x.shape[0]):
                assert np.allclose(
                    seq_x[i], x[i : i + back]
                ), "seq_x samples are not correct"
                assert np.allclose(
                    seq_y[i], y[i + back + forward[0] - 1]
                ), "seq_y samples are not correct"

    def test_numpy_diff(self):
        # Parameters
        features = 5
        length = 500
        back = np.random.randint(2, 50)
        x = np.outer(np.linspace(1, length, length), np.ones(features))
        y = np.linspace(0, length - 1, length)

        # Iterate Scenarios
        for forward in [[1], [5]]:
            # Sequence
            sequence = Sequencer(back=back, forward=forward, diff="diff")
            seq_x, seq_y = sequence.convert(x, y, flat=False)

            # Tests
            assert len(seq_x.shape) == 3, "seq_x is not a tensor"
            assert seq_x.shape[1] == back - 1, "seq_x step dimensions incorrect"
            assert seq_x.shape[2] == features, "seq_x feature dimensions incorrect"
            assert (
                seq_x.shape[0] == seq_y.shape[0]
            ), "seq_x and seq_y inconsistent samples"
            assert np.allclose(seq_x, np.ones_like(seq_x)), "samples seq_x incorrect"
            assert np.allclose(
                seq_y, np.ones_like(seq_y) * forward[0]
            ), "seq_y samples incorrect"

    def test_numpy_log(self):
        # Parameters
        features = 5
        length = 500
        back = np.random.randint(2, 50)
        x = np.outer(np.linspace(1, length, length), np.ones(features))
        y = np.linspace(1, length, length)

        # Iterate Scenarios
        for forward in [[1], [5]]:
            # Sequence
            sequence = Sequencer(back=back, forward=forward, diff="log_diff")
            seq_x, seq_y = sequence.convert(x, y, flat=False)

            # Tests
            assert len(seq_x.shape) == 3, "seq_x is not a tensor"
            assert seq_x.shape[1] == back - 1, "seq_x step dimensions incorrect"
            assert seq_x.shape[2] == features, "seq_x feature dimensions incorrect"
            assert (
                seq_x.shape[0] == seq_y.shape[0]
            ), "seq_x and seq_y inconsistent samples"
            for i in range(seq_x.shape[0]):
                assert np.allclose(
                    seq_x[i], np.log(x[1 + i : i + back]) - np.log(x[i : i + back - 1])
                ), "samples seq_x incorrect"
                assert np.allclose(
                    seq_y[i],
                    np.log(y[i + back + forward[0] - 1]) - np.log(y[i + back - 1]),
                ), "seq_y samples incorrect"

    def test_numpy_multi_out(self):
        # Parameters
        features = 5
        length = 500
        forward = [np.random.randint(2, 10)]
        forward += [forward[0] + np.random.randint(1, 10)]
        back = np.random.randint(2, 50)
        x = np.outer(np.linspace(1, length, length), np.ones(features))
        y = np.linspace(1, length, length)

        # Without differencing
        sequence = Sequencer(back=back, forward=forward, diff="none")
        seq_x, seq_y = sequence.convert(x, y, flat=False)
        # Test
        assert (
            seq_x.shape[0] == seq_y.shape[0]
        ), "seq_x and seq_y have inconsistent samples"
        assert seq_y.shape[1] == 2, "seq_y has incorrect steps"
        for i in range(seq_x.shape[0]):
            assert np.allclose(
                seq_y[i].tolist(),
                [y[i + back + forward[0] - 1], y[i + back + forward[1] - 1]],
            ), "seq_y samples are not correct"

        # With differencing
        sequence = Sequencer(back=back, forward=forward, diff="diff")
        seq_x, seq_y = sequence.convert(x, y, flat=False)
        revert = sequence.revert(seq_y, y[back - 1 : back - 1 + max(forward)])
        # Tests
        assert (
            seq_x.shape[0] == seq_y.shape[0]
        ), "seq_x and seq_y have inconsistent samples"
        assert seq_y.shape[1] == 2, "seq_y has incorrect steps"
        assert np.allclose(
            revert[0, : forward[0] - forward[-1]], y[back - 1 : forward[0] - forward[1]]
        )
        assert np.allclose(revert[1], y[back - 1 :])

    def test_pandas_tensor(self):
        # Parameters
        features = 5
        length = 500
        back = np.random.randint(1, 50)
        x, y = np.random.randint(0, 50, (length, features)), np.random.randint(
            0, 50, length
        )
        x, y = pd.DataFrame(x), pd.Series(y)

        # Iterate scenarios
        for forward in [[1], [5]]:
            # Sequence
            sequence = Sequencer(back=back, forward=forward)
            seq_x, seq_y = sequence.convert(x, y, flat=False)

            # Tests
            assert len(seq_x.shape) == 3, "seq_x is no tensor"
            assert seq_x.shape[1] == back, "seq_x step dimension incorrect"
            assert seq_x.shape[2] == features, "seq_x feature dimension incorrect"
            assert (
                seq_x.shape[0] == seq_y.shape[0]
            ), "seq_x and seq_y have inconsistent samples"
            for i in range(seq_x.shape[0]):
                assert np.allclose(seq_x[i], x[i : i + back]), "seq_x samples incorrect"
                assert np.allclose(
                    seq_y[i], y[i + back + forward[0] - 1]
                ), "seq_y samples incorrect"

    def test_reconstruction(self):
        # Parameters
        length = 100
        features = 5
        back = np.random.randint(2, 10)
        forward = [np.random.randint(2, 10)]
        # x, y = np.random.randint(0, 50, (length, features)), np.random.randint(0, 50, length)  # noqa: E501
        x, y = np.outer(np.linspace(1, length, length), np.ones(features)), np.linspace(
            1, length, length
        )

        # Iterate scenarios
        for diff in ["diff", "log_diff"]:
            # Sequence
            seq = Sequencer(back=back, forward=forward, diff=diff)
            seq_x, seq_y = seq.convert(x, y, flat=False)

            # Tests
            assert len(seq_x.shape) == 3, "seq_x is not a tensor"
            assert seq_x.shape[1] == back - 1, "seq_x step dimensions incorrect"
            assert seq_x.shape[2] == features, "seq_x feature dimensions incorrect"
            assert (
                seq_x.shape[0] == seq_y.shape[0]
            ), "seq_x and seq_y inconsistent samples"
            revert = seq.revert(seq_y, y[back - 1 : back - 1 + forward[0]])
            assert np.allclose(revert, y[back - 1 :]), "reverted seq_y incorrect"



#  Copyright (c) 2022 by Amplo.

import copy
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from sklearn import metrics
from sklearn.model_selection import StratifiedKFold

from amplo.documenting.binary import BinaryDocumenting
from amplo.utils.logging import logger
from amplo.utils.util import deprecated

if TYPE_CHECKING:
    from amplo import Pipeline


@deprecated(
    "This is no longer used, all information is available in the created settings.json."
)
class MultiDocumenting(BinaryDocumenting):
    def __init__(self, pipeline: "Pipeline"):
        super().__init__(pipeline)

    def analyse(self):
        # Initiating
        f1_score = np.zeros((self.p.cv_splits, self.p.n_classes))
        log_loss = np.zeros(self.p.cv_splits)
        avg_acc = np.zeros(self.p.cv_splits)
        cm = np.zeros((self.p.cv_splits, self.p.n_classes, self.p.n_classes))

        # Modelling
        model = None
        self.cv = StratifiedKFold(n_splits=self.p.cv_splits, shuffle=self.p.shuffle)
        for i, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            xt, xv, yt, yv = (
                self.x[t],
                self.x[v],
                self.y[t].reshape((-1)),
                self.y[v].reshape((-1)),
            )
            model = copy.copy(self.model)
            model.fit(xt, yt)
            predictions = model.predict(xv).reshape((-1))
            cm[i] = metrics.confusion_matrix(predictions, yv)

            # Metrics
            f1_score[i] = metrics.f1_score(yv, predictions, average=None)
            avg_acc[i] = metrics.accuracy_score(yv, predictions)
            if hasattr(model, "predict_proba"):
                probabilities = model.predict_proba(xv)
                log_loss[i] = metrics.log_loss(yv, probabilities)

        # Result statistics
        totals = np.sum(cm, axis=(1, 2), keepdims=True)
        means = np.mean(cm / totals * 100, axis=0)
        stds = np.std(cm / totals * 100, axis=0)

        # Store
        self.metrics = {
            "F1 Score": [np.mean(f1_score), np.std(f1_score)],
            "Accuracy": [np.mean(avg_acc), np.std(avg_acc)],
        }
        self.outputMetrics = copy.deepcopy(self.metrics)
        self.outputMetrics["Confusion Matrix Means"] = means
        self.outputMetrics["Confusion Matrix Stds"] = stds
        self.confusion_matrix = {
            "means": means,
            "stds": stds,
        }

        # Print
        logger.info("F1 scores:")
        logger.info("".join([" Class {} |".format(i) for i in range(self.p.n_classes)]))
        logger.info(
            "".join(
                [
                    " {:.2f} % ".ljust(11).format(f1) + "|"
                    for f1 in np.mean(f1_score, axis=0)
                ]
            )
        )
        logger.info(
            "Average Accuracy: {:.2f} \u00B1 {:.2f} %".format(
                np.mean(avg_acc), np.std(avg_acc)
            )
        )
        if hasattr(model, "predict_proba"):
            logger.info(
                "Log Loss:         {:.2f} \u00B1 {:.2f}".format(
                    np.mean(log_loss), np.std(log_loss)
                )
            )
            self.metrics["Log Loss"] = [np.mean(log_loss), np.std(log_loss)]

        if not os.path.exists(
            self.p.main_dir + "EDA/Features/v{}/RF.png".format(self.p.version)
        ):
            if not os.path.exists(
                self.p.main_dir + "EDA/Features/v{}".format(self.p.version)
            ):
                os.makedirs(
                    self.p.main_dir + "EDA/Features/v{}/".format(self.p.version)
                )
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines["right"].set_visible(False)
            ax.spines["bottom"].set_visible(False)
            ax.spines["top"].set_visible(False)
            fi = self.p.feature_processor.feature_importance_["rf"]
            plt.barh(list(fi)[:15], width=list(fi.values())[:15], color="#2369ec")
            fig.savefig(
                self.p.main_dir + "EDA/Features/v{}/RF.png".format(self.p.version),
                format="png",
                dpi=200,
            )

    def model_performance(self):
        self.ln(self.lh)
        self.add_h2("Model Performance")
        self.add_text(
            f"Model performance is analysed by various metrics. This model has been "
            f"selected based on the {self.p.objective} "
            f"score."
        )

        # Metrics
        self.set_font("Helvetica", "B", 12)
        self.ln(self.lh)
        self.cell(w=50, h=self.lh, txt="Metric", border="B", align="C")
        self.cell(w=50, h=self.lh, txt="Score", border="LB", align="C")
        self.set_font("Helvetica", "", 12)
        for k, v in self.metrics.items():
            self.ln(self.lh)
            self.cell(w=50, h=self.lh, txt=k, border="R", align="L")
            self.cell(
                w=50,
                h=self.lh,
                txt="{:.2f} \u00B1 {:.2f} %".format(v[0], v[1]),
                border="L",
                align="C",
            )
        self.ln(self.lh * 3)

        # Confusion Matrix
        nc = self.p.n_classes
        cell_width = int((self.WIDTH - self.pm * 2) / (nc + 2))
        self.add_h3("Confusion Matrix")

        # First row
        self.set_font("Helvetica", "B", 12)
        self.cell(w=cell_width * 2, h=self.lh, txt="", align="L", border="R")
        self.cell(w=cell_width * nc, h=self.lh, txt="True Class", align="C")
        self.ln(self.lh)

        # Second Row
        self.cell(w=cell_width * 2, h=self.lh, txt="", align="L", border="B")
        for i in range(nc):
            self.cell(
                w=cell_width,
                h=self.lh,
                txt="Class {}".format(i),
                align="C",
                border="BL",
            )
        self.ln(self.lh)

        # Values
        for i in range(nc):
            # Set bold
            self.set_font("Helvetica", "B", 12)
            # Prediction (only for first time)
            if i == 0:
                self.cell(w=cell_width, h=self.lh * nc, txt="Prediction", align="L")
            else:
                self.cell(w=cell_width, h=self.lh, txt="")
            # Class
            self.cell(
                w=cell_width, h=self.lh, txt="Class {}".format(i), align="L", border="R"
            )
            # Set normal
            self.set_font("Helvetica", "", 12)
            for j in range(nc):
                self.cell(
                    w=cell_width,
                    h=self.lh,
                    txt="{:.1f}\u00B1{:.1f} %".format(
                        self.confusion_matrix["means"][i][j],
                        self.confusion_matrix["stds"][i][j],
                    ),
                    align="C",
                )
            self.ln(self.lh)

    def validation(self):
        self.ln(self.lh)
        self.check_new_page()
        self.add_h3("Validation Strategy")
        self.add_text(
            "All experiments are cross-validated. This means that every time a model's "
            "performance is evaluated, it's trained on one part of the data, and test "
            "on another. Therefore, the model is always test against data it has not "
            "yet been trained for. This gives the best approximation for real world "
            "(out of sample) performance. The current validation strategy used "
            "is {}, with {} splits and {} shuffling the data.".format(
                type(self.cv).__name__,
                self.p.cv_splits,
                "with" if self.p.shuffle else "without",
            )
        )



#  Copyright (c) 2022 by Amplo.

import copy
import itertools
import random
import re
import time
from datetime import datetime
from typing import Dict, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from amplo.grid_search._base import BaseGridSearch


class ExhaustiveGridSearch(BaseGridSearch):
    """
    Basic exhaustive grid search.

    Parameters
    ----------
    model : Amplo.AutoML.Modeller.ModelType
        Model object to optimize.
    params : optional
        Parameters to optimize. Has no effect for `OptunaGridSearch`.
    n_trials : int
        Limit the number of trials/candidates to search.
    timeout : int
        Limit the time for optimization.
    cv : sklearn.model_selection.BaseCrossValidator
        Cross validation object.
    scoring : str or sklearn.metrics._scorer._BaseScorer
        A valid string for `sklearn.metrics.SCORERS`
    verbose : int
        Verbose logging.
    """

    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, **kwargs)

        # Initiate
        self.parsed_params = []
        self.result = []

    def _get_hyper_params(self) -> Dict[str, Union[None, bool, int, float, str, list]]:
        param_values = self._hyper_parameter_values
        # Drop conditionals as they are not supported
        param_values.pop("CONDITIONALS", None)

        # Drop/adjust some more parameter names which HalvingGridSearch cannot handle
        model_name = type(self.model).__name__
        if model_name in ("BaggingClassifier", "BaggingRegressor"):
            param_values["max_features"] = ("categorical", [0.5, 0.8, 0.9, 1.0], None)
        elif model_name in ("LGBMRegressor", "LGBMClassifier"):
            # Someway it cannot handle uniform distributions for this two cases
            param_values["bagging_fraction"] = (
                "categorical",
                [0.5, 0.7, 0.9, 1.0],
                None,
            )
            param_values["feature_fraction"] = (
                "categorical",
                [0.5, 0.7, 0.9, 1.0],
                None,
            )

        # Extract parameter distributions
        params = {}
        for p_name, value in param_values.items():

            # Read out
            p_type = value[0]  # parameter type (str)
            p_args = value[1]  # parameter arguments (list, tuple)
            p_count = (
                value[2] if p_type != "categorical" else None
            )  # n_samples for np.linspace or np.logspace

            # Sanity checks
            assert (
                len(p_args) == 2 or p_type == "categorical"
            ), "Only categorical parameter can have more/less than two suggest args"

            # Skip parameters when none are wanted
            if p_count == 0:
                continue
            # Suggest parameter given the arguments
            if p_type == "categorical":
                params[p_name] = p_args
                continue
            elif p_type in ("int", "uniform"):
                params[p_name] = np.linspace(*p_args, p_count)
            elif p_type in ("logint", "loguniform"):
                p_args = [np.log(arg) for arg in p_args]  # logarithmise for np.logspace
                params[p_name] = np.logspace(*p_args, p_count)
            else:
                raise NotImplementedError("Invalid parameter specification")
            # Discretize and convert to list
            if "int" in p_type:
                params[p_name] = params[p_name].astype("int")
            params[p_name] = np.unique(params[p_name]).tolist()

        return params

    def _parse_params(self):
        if len(self.params.items()) > 0:
            k, v = zip(*self.params.items())
            self.parsed_params = [
                dict(zip(k, v)) for v in itertools.product(*self.params.values())
            ]
            random.shuffle(self.parsed_params)
            self.logger.info(
                f"[GridSearch] Optimizing {type(self.model).__name__}, "
                f"{self.cv.n_splits}-folds with {len(self.parsed_params)} parameter "
                f"combinations and {len(self.parsed_params) * self.cv.n_splits} runs."
            )
        else:
            self.parsed_params = [{}]

    def fit(self, x, y):
        start_time = time.time()
        # todo check data
        # Parse Params
        model_name = type(self.model).__name__
        is_classification = bool(
            re.match(r".*(Classification|Classifier|SVC)", model_name)
        )
        if is_classification:
            if len(np.unique(y)) == 2:
                self.binary = True
            else:
                self.binary = False
        # Convert to Numpy
        if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
            x = np.array(x)
        if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            y = np.array(y).reshape((-1))
        self.samples = len(y)

        # Get params
        if self.params is None:
            self.params = self._get_hyper_params()
        self._parse_params()

        # Loop through parameters
        for i, param in tqdm(enumerate(self.parsed_params)):
            # Timeout or trials limit
            if time.time() - start_time > self.timeout or i > self.n_trials:
                return pd.DataFrame(self.result)

            # Scoring and timing vector
            scoring = []
            timing = []

            # Cross Validation
            for train_ind, val_ind in self.cv.split(x, y):
                # Start Timer
                t = time.time()

                # Split data
                x_train, x_val = x[train_ind], x[val_ind]
                y_train, y_val = y[train_ind], y[val_ind]

                # Model training
                model = copy.deepcopy(self.model)
                model.set_params(**param)
                model.fit(x_train, y_train)

                # Results
                scoring.append(self.scoring(model, x_val, y_val))
                timing.append(time.time() - t)

            # Output Printing
            self.logger.info(
                f"[GridSearch][{datetime.now().strftime('%H:%M')}] "
                f"Score: {np.mean(scoring):.4f} \u00B1 {np.std(scoring):.4f} "
                f"(in {np.mean(timing):.1f} seconds)."
            )

            self.result.append(
                {
                    "date": datetime.today().strftime("%d %b %y"),
                    "model": type(model).__name__,
                    "mean_objective": np.mean(scoring),
                    "std_objective": np.std(scoring),
                    "worst_case": np.mean(scoring) - np.std(scoring),
                    "params": param,
                    "mean_time": np.mean(timing),
                    "std_time": np.std(timing),
                }
            )
        return pd.DataFrame(self.result)



#  Copyright (c) 2022 by Amplo.

from amplo.observation._base import ProductionWarning
from amplo.observation.data import DataObserver
from amplo.observation.model import ModelObserver
from amplo.observation.production import ProductionObserver



#  Copyright (c) 2022 by Amplo.

import pytest

from amplo.api.platform import PlatformSynchronizer
from amplo.utils.testing import make_production_data
from tests.unit.api import TestAPI


class TestPlatform(TestAPI):
    def test_upload(self):
        """
        This only tests that upload_latest_model doesn't error.
        - Good to check whether the file exists in the cloud.
        - Also needs some cleanup to make sure it doesn't stay there.
        - Also does upload_latest_model truly take the latest?
        - What if latest locally and in the cloud are outdated? It needs to go in the right folder.
        """
        # TODO: implement as soon as platform is out of beta mode
        pytest.skip("Test not yet implemented")

        # Make dummy production data
        issue_dir, kwargs = make_production_data(
            self.sync_dir, team="Demo", machine="Charger 75kW", service="Diagnostics"
        )
        # Upload
        sync = PlatformSynchronizer()
        sync.upload_latest_model(issue_dir, **kwargs)



#  Copyright (c) 2022 by Amplo.
import xgboost.callback
from sklearn.model_selection import train_test_split
from xgboost import XGBClassifier as _XGBClassifier

from amplo.classification._base import BaseClassifier
from amplo.utils import check_dtypes


def _validate_xgboost_callbacks(callbacks):
    if not callbacks:
        return []

    for cb in callbacks:
        raise NotImplementedError

    return callbacks


class XGBClassifier(BaseClassifier):
    """
    Amplo wrapper for xgboost.XGBClassifier.

    Parameters
    ----------
    callbacks : list of str, optional
        ...
    test_size : float, default: 0.1
        Test size for train-test-split in fitting the model.
    early_stopping_rounds : int, default: 100
        Number of early stopping rounds.
    random_state : int, default: None
        Random state for train-test-split in fitting the model.
    verbose : {0, 1, 2}, default: 0
        Verbose logging.
    **model_params : Any
        Model parameters for underlying xgboost.XGBClassifier.
    """

    model: _XGBClassifier  # type hint

    def __init__(
        self,
        callbacks=None,
        test_size=0.1,
        early_stopping_rounds=None,
        random_state=None,
        verbose=0,
        **model_params,
    ):
        # Verify input dtypes and integrity
        check_dtypes(
            ("callbacks", callbacks, (type(None), list)),
            ("test_size", test_size, float),
            ("random_state", random_state, (type(None), int)),
            ("early_stopping_rounds", early_stopping_rounds, (type(None), int)),
            ("model_params", model_params, dict),
        )
        if not 0 <= test_size < 1:
            raise ValueError(f"Invalid attribute for test_size: {test_size}")

        # Set attributes
        self.callbacks = callbacks
        self.test_size = test_size
        self.early_stopping_rounds = early_stopping_rounds or 100
        self.random_state = random_state

        # Set up model
        default_model_params = {
            "n_estimators": 100,  # number of boosting rounds
            "random_state": random_state,
            "use_label_encoder": False,  # is deprecated
            "verbosity": verbose,
        }
        for k, v in default_model_params.items():
            if k not in model_params:
                model_params[k] = v
        model = _XGBClassifier(**model_params)

        super().__init__(model=model, verbose=verbose)

    def _fit(self, x, y=None, **fit_params):
        # Set up fitting callbacks
        callbacks = _validate_xgboost_callbacks(self.callbacks)
        callbacks.append(xgboost.callback.EarlyStopping(self.early_stopping_rounds))

        # Split data and fit model
        xt, xv, yt, yv = train_test_split(
            x, y, stratify=y, test_size=self.test_size, random_state=self.random_state
        )
        self.model.fit(
            xt, yt, eval_set=[(xv, yv)], callbacks=callbacks, verbose=bool(self.verbose)
        )
